#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/__init__.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.06.2020
# Last Modified Date:  11.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

# flake8: noqa: F401
from .AmpelLogger import DEBUG, ERROR, INFO, SHOUT, VERBOSE, WARNING, AmpelLogger
from .LogFlag import LogFlag
